#include "lcd.h"

#include <avr/interrupt.h>
#include <avr/io.h>

#include <util/delay.h>

#include <stdio.h>
#include <math.h>

#define SENSOR_TRIG_TIME 10 // uS
#define SENSOR_WAIT_TIME 20 // mS + timp lcd

#define SENSOR_DDR DDRC
#define SENSOR_PORT PORTC
#define SENSOR_PIN PINC

#define SENSOR_ECHO PC0
#define SENSOR_TRIG PC1

#define SENSOR_INT_VAL PCIE2
#define SENSOR_INT_VECTOR PCINT2_vect
#define SENSOR_INT_MSK PCMSK2
#define SENSOR_INT_MSK_VAL PCINT16

#define SENSOR_TIMER_CNT TCNT1
#define SENSOR_TIMER_PORT_A TCCR1A
#define SENSOR_TIMER_PORT_B TCCR1B
#define SENSOR_TIMER_PORT_INT TIMSK1


#define MOTORS_DDR DDRD
#define MOTORS_PORT PORTD

// not used
#define LMOTOR_PWM PD7
#define RMOTOR_PWM PD6

// am conectat motoarele invers
#define LMOTOR_EN PD3
#define LMOTOR_EN1 PD2
#define RMOTOR_EN1 PD1
#define RMOTOR_EN PD0   

#define MOTORS_TIMER_PORT_A TCCR2A
#define MOTORS_TIMER_PORT_B TCCR2B
#define MOTORS_TIMER_PORT_INT TIMSK2
#define LMOTOR_PWM_REG OCR2A
#define RMOTOR_PWM_REG OCR2B
#define MIN_PWM 175
#define MAX_PWM 225
#define MAX_SPEED_DIF_SCALING 1.8


#define SOUND_SPEED_CM_uS 0.0343
#define NEXT_STATE_READS 3
#define MAX_TURN_TIMES 2
#define TRIGGER_BACK_MOVE 6 // cm
#define DELAY_BACK_MOVE 400 // ms

enum MOVE_STATE {
    STOP = 0,
    FAST_TURN = 1,
    NORMAL_TURN = 2,
    SLOW_TURN = 3,
    NORMAL = 4,
    SPEED = 5
} move_state, last_mesured_move_state;

// citiri necesare pana la schimbare stare
uint8_t go_to_next_state;

uint16_t distance;
/*
 * ultimul element e pentru calcularea vitezei
 * la starea SPEED (viteza max la 100cm distanta)
 * la starile SPEED si NORMAL viteza se calculeaza ca
 * (distanta / distanta_max_stare) * vit. max
 * la celelalte este constanta egala cu vit. max
 */
uint8_t move_state_threshold[6] = {0, 18, 35, 60, 75, 90};
// % din viteza maxima
uint8_t move_state_max_speed[6] = {0, 45, 65, 70, 85, 100};
/*
 * diferenta minima de viteza intre motoare la viraje egala cu
 * % din diferenta maxima intre motoare -> left_s = -right_s -> 2 * speed
 * diferenta efectiva va varia si cu procentul distantei din threshold
 * primul element e ca sa pot accesa [move_state]
 */
uint8_t turning_speed_min_dif[4] = {0, 90, 45, 25};

// % din viteza maxima
int8_t lmotor_speed, rmotor_speed;
// >=0 stanga, <0 dreapta
int8_t turn_dir;
// bool pastrare directie viraj
uint8_t was_turning;

void inline print_data_on_lcd() {
    char str[14];
    
    LCD_writeInstr(LCD_INSTR_clearDisplay);
    sprintf(str, "T %hu %d cm", SENSOR_TIMER_CNT, distance);
    LCD_printAt(0x0, str);
    sprintf(str, "l %hhd r %hhd S%d", lmotor_speed, rmotor_speed, move_state);
    LCD_printAt(0x40, str);
}

void inline set_motor_direction(uint8_t en, uint8_t en1, int8_t speed) {
    if (speed >= 0) {
        MOTORS_PORT |= (1 << en);
        MOTORS_PORT &= ~(1 << en1);
    } else {
        MOTORS_PORT &= ~(1 << en);
        MOTORS_PORT |= (1 << en1);
    }
}


void init_motors_port() {
    // output pe toti pinii catre driver
    MOTORS_DDR |= (1 << LMOTOR_PWM | 1 << RMOTOR_PWM |
                1 << LMOTOR_EN | 1 << LMOTOR_EN1 |
                1 << RMOTOR_EN | 1 << RMOTOR_EN1);

    // clear la pinii de pwm
    MOTORS_PORT &= ~(1 << LMOTOR_PWM | 1 << RMOTOR_PWM);

    // setare directie in fata la motoare
    set_motor_direction(LMOTOR_EN, LMOTOR_EN1, 1);
    set_motor_direction(RMOTOR_EN, RMOTOR_EN1, 1);
}

void init_move() {
    move_state = STOP;
    go_to_next_state = NEXT_STATE_READS;

    LMOTOR_PWM_REG = 0;
    RMOTOR_PWM_REG = 0;

    turn_dir = MAX_TURN_TIMES;
    was_turning = 0;
}

void init_motors_timer() {
    // no int
    MOTORS_TIMER_PORT_INT = 0;
    // fastpwm - clear on comp, set on bottom
    MOTORS_TIMER_PORT_A = (1 << COM2A1) | (1 << COM2B1) |
                    (1 << WGM21) | (1 << WGM20);
    // prescaler 1
    MOTORS_TIMER_PORT_B = (1 << CS20);
}


void init_echo_trig_port() {
    SENSOR_DDR &= ~(1 << SENSOR_ECHO);
    SENSOR_DDR |= (1 << SENSOR_TRIG);

    // nu vreau rezistenta de pull-up pe ECHO
    SENSOR_PORT &= ~(1 << SENSOR_TRIG | 1 << SENSOR_ECHO);
}

void init_int_echo() {
    SENSOR_INT_MSK = (1 << SENSOR_INT_MSK_VAL);
    PCICR = (1 << SENSOR_INT_VAL);

    sei();
}

inline void reset_sensor_timer() {
    SENSOR_TIMER_CNT = 0;
}

inline void init_sensor_timer() {
    SENSOR_TIMER_PORT_INT = 0; // no int
    SENSOR_TIMER_PORT_A = 0; // mod normal
    // mod normal, prescaler 8 -> frecv 2MHz
    SENSOR_TIMER_PORT_B = (1 << CS11);

    reset_sensor_timer();
}

inline void stop_sensor_timer() {
    SENSOR_TIMER_PORT_B = 0;
}


inline enum MOVE_STATE get_move_state() {
    uint8_t i;

    // incadrez distanta intr-un din starile de miscare
    for (i = 0; i < 4; ++i)
        if (move_state_threshold[i] <= distance &&
                distance < move_state_threshold[i + 1])
            break;

    switch (i) {
        case 0: return FAST_TURN;
        case 1: return NORMAL_TURN;
        case 2: return SLOW_TURN;
        case 3: return NORMAL;
    }

    // daca nu s-a incadrat in primele 4
    return SPEED;
}

inline void modify_move_state() {
    // daca starea citita este diferita de cea curenta
    if (move_state != last_mesured_move_state) {
        // scade numarul de citiri necesare pana la schimbare stare
        --go_to_next_state;
        // daca este 0 schimb starea
        if (go_to_next_state == 0) {
            move_state = last_mesured_move_state;
            go_to_next_state = NEXT_STATE_READS;
        }
    } else {
        // altfel resetez numarul de citiri
        go_to_next_state = NEXT_STATE_READS;
    }
}

inline void set_rotate_speed(int8_t *princm, int8_t *seconm, float percent, int8_t speed) {
    float percent_speed_dif;
    uint16_t speed_dif;

    *princm = speed;
    
    // procentul diferentei de viteza dintre motoare
    percent_speed_dif = (float) turning_speed_min_dif[move_state] / 100;
    // scalarea nu poate fi mai mare de MAX_SPEED_DIF_SCALING
    if (percent < MAX_SPEED_DIF_SCALING)
        percent_speed_dif *= percent;

    // diferenta max e 2 * speed
    speed_dif = (2 * speed);
    if (percent_speed_dif < 1)
        speed_dif *= percent_speed_dif;
    
    *seconm = speed - speed_dif;
}

inline void calculate_new_speed() {
    float percent;

    // procentul din viteza maxima este calculat
    // in functie de distanta maxima din threshold
    percent = (float) distance / move_state_threshold[move_state];
    
    switch (move_state) {
        case SPEED:
        case NORMAL: {
            lmotor_speed = move_state_max_speed[move_state];

            // la starea SPEED distanta poate fi
            // mai mare decat cea din threshold
            if (percent < 1) {
                lmotor_speed = percent * lmotor_speed; 
            }

            // motoarele au aceeasi viteza
            rmotor_speed = lmotor_speed;
            was_turning = 0;
            break;
        }
        default: {
            // daca nu este deja in viraj
            if (!was_turning) {
                // daca am ajuns la numarul maxim de viraje dreapta
                if (turn_dir == -MAX_TURN_TIMES)
                    turn_dir = MAX_TURN_TIMES;

                --turn_dir;
                was_turning = 1;
            }

            // procentul de scalare este invers fata de cel al starilor fara viraj
            percent = 1 + (1 - percent);

            // daca suntem intr-un viraj stanga
            if (turn_dir >= 0)
                // motorul drept este cel principal
                set_rotate_speed(&rmotor_speed, &lmotor_speed,
                        percent, move_state_max_speed[move_state]);
            else
                // motorul stang este cel principal
                set_rotate_speed(&lmotor_speed, &rmotor_speed,
                        percent, move_state_max_speed[move_state]);
        }
    }
}

inline void adjust_pwm() {
    float PWM;
    uint8_t dif = MAX_PWM - MIN_PWM;

    // viteza este % din diferenta intre MAX_PWM si MIN_PWM
    if (lmotor_speed == 0) {
        LMOTOR_PWM_REG = 0;
    } else {
        PWM = fabs(lmotor_speed) / 100 * dif;
        LMOTOR_PWM_REG = MIN_PWM + (uint8_t) PWM;

        // seteaza directia motorului
        set_motor_direction(LMOTOR_EN, LMOTOR_EN1, lmotor_speed);
    }

    if (rmotor_speed == 0) {
        RMOTOR_PWM_REG = 0;
    } else {
        PWM = fabs(rmotor_speed) / 100 * dif;
        RMOTOR_PWM_REG = MIN_PWM + (uint8_t) PWM;

        set_motor_direction(RMOTOR_EN, RMOTOR_EN1, rmotor_speed);
    }
}

inline void go_back() {
    lmotor_speed = -100;
    rmotor_speed = -100;

    adjust_pwm();

    print_data_on_lcd();
    _delay_ms(DELAY_BACK_MOVE);
}

inline void decide() {
    // citeste noua stare
    last_mesured_move_state = get_move_state();
    // verifica daca trebuie modificata starea curenta
    modify_move_state();

    // daca distanta e prea mica mergi inapoi
    if (move_state == FAST_TURN && distance <= TRIGGER_BACK_MOVE)
        go_back();

    // calculeaza noua viteza - aceasta va varia in functie de distanta
    calculate_new_speed();
    adjust_pwm();
}

ISR(SENSOR_INT_VECTOR) {
    // protectie pentru echo fals
    static uint8_t echo_state = 0;

    if (SENSOR_PIN & (1 << SENSOR_ECHO)) {
        if (echo_state == 0) {
            // echo -> start timer
            init_sensor_timer();
            echo_state = 1;
        }
    } else if (echo_state == 1) {
        // echo s-a terminat -> stop timer
        stop_sensor_timer();
        echo_state = 0;

        // impart la 2 pentru ca frecv timer-ului e 2MHz
        // si inca odata pentru ca unda face distanta de doua ori
        distance = (uint16_t)ceil(SENSOR_TIMER_CNT * SOUND_SPEED_CM_uS / 2 / 2);
        decide();

        // lcd print
        print_data_on_lcd();
    }
}

void main_loop()
{
    // main loop
    for(;;) {
        // trigger senzor
        SENSOR_PORT |= (1 << SENSOR_TRIG);
        _delay_us(SENSOR_TRIG_TIME);
        SENSOR_PORT &= ~(1 << SENSOR_TRIG);

        // asteapta echo -> si decizia
        _delay_ms(SENSOR_WAIT_TIME);
    }
}

int main()
{
    LCD_init();

    init_motors_port();
    init_move();
    init_motors_timer();

    init_echo_trig_port();
    init_int_echo();

    // astept 2 sec inainte de start
    LCD_printAt(0x00, "Start in 2s");
    _delay_ms(2000);

    main_loop();
}
